"""_fetchrepodata: Map yum/dnf repo metadata to local lookup caches"""
import gzip
import json
import logging
import os
from collections import defaultdict

import click
import requests

import gi
gi.require_version('Modulemd', '1.0')
from gi.repository import Modulemd

from attr import attributes, attrib
from lxml import etree
from requests_toolbelt.downloadutils.tee import tee_to_file
from urllib.parse import urljoin

XDG_CACHE_HOME = os.environ.get("XDG_CACHE_HOME") or os.path.expanduser("~/.cache")
CACHEDIR = os.path.join(XDG_CACHE_HOME, "fedmod")

log = logging.getLogger(__name__)

class MissingMetadata(Exception):
    """Reports failure to find the local metadata cache"""

@attributes
class RepoPaths:
    remote_repo_url = attrib(type=str)
    remote_metadata_url = attrib(type=str)
    local_cache_path = attrib(type=str)
    local_metadata_path = attrib(type=str)

@attributes
class RepoPathsPair:
    arch = attrib(type=RepoPaths)
    src = attrib(type=RepoPaths)

def _define_repo(remote_prefix, local_cache_name, arch=None):
    if arch is None:
        local_arch_path = "source"
        if "updates" in remote_prefix:
            remote_arch_path = "SRPMS"
        else:
            remote_arch_path = "source/tree/"
    else:
        local_arch_path = arch
        if "updates" in remote_prefix:
            remote_arch_path = arch + "/"
        else:
            remote_arch_path = os.path.join(arch, "os/")
    remote_repo_url = os.path.join(remote_prefix, remote_arch_path)
    remote_metadata_url = os.path.join(remote_repo_url, "repodata/")
    local_cache_path = os.path.join(CACHEDIR, "repos", local_cache_name, local_arch_path)
    local_metadata_path = os.path.join(local_cache_path, "repodata/")
    return RepoPaths(remote_repo_url, remote_metadata_url,
                     local_cache_path, local_metadata_path)

def _define_repos(remote_prefix, local_cache_name, arch):
    return RepoPathsPair(arch=_define_repo(remote_prefix, local_cache_name, arch),
                         src=_define_repo(remote_prefix, local_cache_name))

@attributes
class DistroRelease:
    name = attrib(type=str)
    main_repo = attrib(type=str)
    updates_repo = attrib(default=None, type=str)
    modular_updates_repo = attrib(default=None, type=str)
    modular_repo = attrib(default=None, type=str)

class DistroPaths(object):
    def __init__(self, release, arch):
        self.release = release
        self.arch = arch

        self.main = _define_repos(release.main_repo, release.name + "-packages", arch)

        if release.updates_repo:
            self.updates = _define_repos(release.updates_repo, release.name + "-updates", arch)
        else:
            self.updates = None

        if release.modular_repo:
            self.modular = _define_repos(release.modular_repo, release.name + "-modular", arch)
        else:
            self.modular = None

        if release.modular_updates_repo:
            self.modular_updates = _define_repos(release.modular_updates_repo, release.name + "-modular-updates", arch)
        else:
            self.modular_updates = None

        self.all_repos = []
        for pair in (self.main, self.updates, self.modular, self.modular_updates):
            if pair:
                self.all_repos.append(pair.arch)
                self.all_repos.append(pair.src)

_DISTRO_RELEASES = {
    "f28": DistroRelease(
        name="f28",
        main_repo="https://dl.fedoraproject.org/pub/fedora/linux/releases/28/Everything/",
        updates_repo="https://dl.fedoraproject.org/pub/fedora/linux/updates/28/Everything/",
        modular_repo="https://dl.fedoraproject.org/pub/fedora/linux/releases/28/Modular/",
        modular_updates_repo="https://dl.fedoraproject.org/pub/fedora/linux/updates/28/Modular/",
    ),
    "f29": DistroRelease(
        name="f29",
        main_repo="https://dl.fedoraproject.org/pub/fedora/linux/development/29/Everything/",
        modular_repo="https://dl.fedoraproject.org/pub/fedora/linux/development/29/Modular/",
    ),
    "rawhide": DistroRelease(
        name="rawhide",
        main_repo="https://dl.fedoraproject.org/pub/fedora/linux/development/rawhide/Everything/",
        modular_repo="https://dl.fedoraproject.org/pub/fedora/linux/development/rawhide/Modular/",
    )
}
_DEFAULT_RELEASE_NAME = "f29"

_ARCHES = set(["aarch64", "armv7hl", "i686", "ppc64", "ppc64le", "s390x", "x86_64"])
_DEFAULT_ARCH = "x86_64"

def parse_dataset_name(dataset_name):
    """Parse dataset_name as distro:release and validate the result"""

    if dataset_name is None:
        return _DEFAULT_RELEASE_NAME, _DEFAULT_ARCH

    parts = dataset_name.split("-")
    if len(parts) == 1:
        release_name, arch = parts[0], _DEFAULT_ARCH
    elif len(parts) == 2:
        release_name, arch = parts[0], parts[1]
    else:
        raise ValueError("dataset name should have the form RELEASE[-ARCH]")

    if release_name not in _DISTRO_RELEASES:
        raise ValueError("Unknown release name: {}. Known releases: {}"
                         .format(release_name,
                                 ", ".join(sorted(_DISTRO_RELEASES.keys()))))

    if arch not in _ARCHES:
        raise ValueError("Unknown architecture: {}. Known releases: {}"
                         .format(arch,
                                 ", ".join(sorted(_ARCHES))))

    return release_name, arch

def _display_dataset_name(dataset_name):
    release_name, arch = parse_dataset_name(dataset_name)
    if release_name == _DEFAULT_RELEASE_NAME and arch == _DEFAULT_ARCH:
        return None
    elif arch == _DEFAULT_ARCH:
        return release_name
    else:
        return dataset_name

def _get_distro_paths(dataset_name):
    release_name, arch = parse_dataset_name(dataset_name)
    return DistroPaths(_DISTRO_RELEASES[release_name], arch)

_MODULE_FORWARD_LOOKUP_CACHE = "module-contents"
_SRPM_REVERSE_LOOKUP_CACHE = "srpm-to-module"
_RPM_REVERSE_LOOKUP_CACHE = "rpm-to-module"

_ALL_CACHES = [_MODULE_FORWARD_LOOKUP_CACHE,
               _SRPM_REVERSE_LOOKUP_CACHE,
               _RPM_REVERSE_LOOKUP_CACHE]

METADATA_SECTIONS = ("filelists", "primary", "modules")

_REPOMD_XML_NAMESPACE = {"rpm": "http://linux.duke.edu/metadata/repo"}
def _read_repomd_location(repomd_xml, section):
    location = repomd_xml.find(f"rpm:data[@type='{section}']/rpm:location", _REPOMD_XML_NAMESPACE)
    if location is not None:
        return location.attrib["href"]
    return None

def _download_one_file(remote_url, filename):
    if os.path.exists(filename) and not filename.endswith((".xml", ".yaml")):
        print(f"  Skipping download; {filename} already exists")
        return
    response = requests.get(remote_url, stream=True)
    try:
        print(f"  Downloading {remote_url}")
        chunksize = 65536
        expected_chunks = int(response.headers["content-length"]) / chunksize
        downloader = tee_to_file(response, filename=filename, chunksize=chunksize)
        show_progress = click.progressbar(downloader, length=expected_chunks)
        with show_progress:
            for chunk in show_progress:
                pass
    finally:
        response.close()
    print(f"  Added {filename} to cache")

def _download_metadata_files(repo_paths):
    local_path = repo_paths.local_cache_path
    local_metadata_path = repo_paths.local_metadata_path
    os.makedirs(local_metadata_path, exist_ok=True)
    repomd_url = urljoin(repo_paths.remote_metadata_url, "repomd.xml")
    print(f"Remote metadata: {repomd_url}")
    response = requests.get(repomd_url)
    response.raise_for_status()
    repomd_filename = os.path.join(local_metadata_path, "repomd.xml")
    with open(repomd_filename, "wb") as f:
        f.write(response.content)
    print(f"  Cached metadata in {repomd_filename}")
    repomd_xml = etree.parse(repomd_filename)
    files_to_fetch = set()
    for section in METADATA_SECTIONS:
        relative_href = _read_repomd_location(repomd_xml, section)
        if relative_href is not None:
            files_to_fetch.add(relative_href)
    predownload = set(os.listdir(local_path))
    for relative_href in files_to_fetch:
        absolute_href = urljoin(repo_paths.remote_repo_url, relative_href)
        filename = os.path.join(local_path, relative_href)
        # This could be parallelised with concurrent.futures, but
        # probably not worth it (it makes the progress bars trickier)
        _download_one_file(absolute_href, filename)
    postdownload = set(os.listdir(local_path))
    # Prune any old metadata files automatically
    if len(postdownload) >= (len(predownload) + len(METADATA_SECTIONS)):
        # TODO: Actually prune old metadata files
        pass

def _cache_fname(paths, cache_name):
    return os.path.join(CACHEDIR, f"{paths.release.name}-{cache_name}-{paths.arch}-cache.json")

def _write_cache(paths, cache_name, data):
    """Write the given data to the nominated cache file"""
    cache_fname = _cache_fname(paths, cache_name)
    with open(cache_fname, "w") as cache_file:
        json.dump(data, cache_file)
    print(f"  Added {cache_fname} to cache")

def _read_cache(paths, cache_name):
    """Read the parsed data from the nominated cache file"""
    cache_fname = _cache_fname(paths, cache_name)
    with open(cache_fname, "r") as cache_file:
        return json.load(cache_file)

def _read_modules(repo_paths):
    metadata_dir = os.path.join(repo_paths.local_cache_path)
    repomd_fname = os.path.join(metadata_dir, "repodata", "repomd.xml")
    repomd_xml = etree.parse(repomd_fname)
    repo_relative_modulemd = _read_repomd_location(repomd_xml, "modules")
    repo_modulemd_fname = os.path.join(metadata_dir, repo_relative_modulemd)
    with gzip.open(repo_modulemd_fname, "rt") as modules_yaml_gz:
        modules_yaml = modules_yaml_gz.read()

    objects = Modulemd.objects_from_string(modules_yaml)
    return (o for o in objects if isinstance(o,  Modulemd.Module))

def _merge_modules(module_sets):
    modules = dict()

    for module_set in module_sets:
        for m in module_set:
            old = modules.get((m.props.name, m.props.stream))
            if not old or m.props.version > old.props.version:
                modules[(m.props.name, m.props.stream)] = m

    return modules.values()

def _write_lookup_caches(paths):
    module_sets = []
    if paths.modular:
        module_sets.append(_read_modules(paths.modular.arch))
    if paths.modular_updates:
        module_sets.append(_read_modules(paths.modular_updates.arch))

    module_forward_lookup = {}
    srpm_reverse_lookup = defaultdict(list)
    rpm_reverse_lookup = defaultdict(list)

    for module in _merge_modules(module_sets):
        module_name = module.props.name
        artifacts = module.props.rpm_artifacts.get()

        module_forward_lookup[module_name] = list(set(artifacts))
        components = module.props.components_rpm
        for srpmname in components:
            srpm_reverse_lookup[srpmname].append(module_name)
        for rpmname in artifacts:
            rpmprefix = rpmname.split(":", 1)[0].rsplit("-", 1)[0]
            rpm_reverse_lookup[rpmprefix].append(module_name)

    # Cache the lookup tables as local JSON files
    print("Caching lookup tables")
    _write_cache(paths, _MODULE_FORWARD_LOOKUP_CACHE, module_forward_lookup)
    _write_cache(paths, _SRPM_REVERSE_LOOKUP_CACHE, srpm_reverse_lookup)
    _write_cache(paths, _RPM_REVERSE_LOOKUP_CACHE, rpm_reverse_lookup)


def download_repo_metadata(dataset_name):
    """Downloads the latest repo metadata"""

    paths = _get_distro_paths(dataset_name)
    for repo_definition in paths.all_repos:
        _download_metadata_files(repo_definition)
    _write_lookup_caches(paths)

@attributes
class LocalMetadataCache:
    dataset_name = attrib(type=str)
    cache_dir = attrib(type=str)
    srpm_to_modules = attrib(type=dict)
    rpm_to_modules = attrib(type=dict)
    module_to_packages = attrib(type=dict)
    source_repo_cache = attrib(type=str)
    arch_repo_cache = attrib(type=str)
    source_updates_cache = attrib(type=str)
    arch_updates_cache = attrib(type=str)


def load_cached_repodata(dataset_name):
    paths = _get_distro_paths(dataset_name)

    display_name = _display_dataset_name(dataset_name)
    arg = f" --dataset={display_name}" if display_name else ""

    # Check whether or not fetch-metadata has been run at all
    metadata_dir = os.path.join(paths.main.arch.local_cache_path)
    repomd_fname = os.path.join(metadata_dir, "repodata", "repomd.xml")

    if not os.path.exists(repomd_fname):
        msg = f"{repomd_fname!r} does not exist. Run `fedmod{arg} fetch-metadata`."
        raise MissingMetadata(msg)
    # Check whether or not fetch-metadata actually finished
    for cache in _ALL_CACHES:
        cache_fname = _cache_fname(paths, cache)
        if not os.path.exists(cache_fname):
            msg = (f"{cache_fname!r} does not exist. "
                   f"Try running `fedmod{arg} fetch-metadata` again.")
            raise MissingMetadata(msg)
    # Load the metadata
    return LocalMetadataCache(
        dataset_name = dataset_name,
        cache_dir = CACHEDIR,
        srpm_to_modules = _read_cache(paths, _SRPM_REVERSE_LOOKUP_CACHE),
        rpm_to_modules = _read_cache(paths, _RPM_REVERSE_LOOKUP_CACHE),
        module_to_packages = _read_cache(paths, _MODULE_FORWARD_LOOKUP_CACHE),
        source_repo_cache = paths.main.src.local_cache_path,
        arch_repo_cache = paths.main.arch.local_cache_path,
        source_updates_cache = paths.updates.src.local_cache_path if paths.updates else None,
        arch_updates_cache = paths.updates.arch.local_cache_path if paths.updates else None,
    )
