import click
import json
import koji
import sys
import logging
import os
from textwrap import dedent

import gi
gi.require_version('Modulemd', '1.0')
from gi.repository import Modulemd

from ._repodata import dataset_release_name
from .get_module_builds import get_module_builds
from .module_generator import ModuleGenerator
from .util import rpm_name_only
from . import _depchase

def dataset_release_branch():
    release_name = dataset_release_name()
    if release_name == 'rawhide':
        return 'master'
    else:
        return release_name


def _get_runtime_packages():
    builds = get_module_builds('flatpak-runtime', dataset_release_branch())
    # Each flatpak-runtime stream should be built against a single context
    assert len(builds) == 1
    build = builds[0]

    mmd_str = build['extra']['typeinfo']['module']['modulemd_str']
    mmd = Modulemd.Module.new_from_string(mmd_str)

    return set(mmd.peek_profiles()['runtime'].peek_rpms().get())


class FlatpakGenerator(ModuleGenerator):
    def __init__(self, pkg):
        super().__init__([pkg])
        self.rpm_branch = dataset_release_branch()

    def _calculate_dependencies(self):
        pkg = self.pkgs[0]
        pool = self._pool

        self.api_srpms = {rpm_name_only(_depchase.get_srpm_for_rpm(pool, pkg))}

        runtime_packages = _get_runtime_packages()
        all_needed_packages = _depchase.ensure_installable(pool, [pkg], hints=runtime_packages)

        pkgs = all_needed_packages - runtime_packages

        run_srpms = {rpm_name_only(_depchase.get_srpm_for_rpm(pool, dep)) for dep in pkgs}
        self.run_srpms = run_srpms - self.api_srpms

        self.module_run_deps = {'flatpak-runtime': [dataset_release_branch()]}

    def _update_module_md(self):
        super()._update_module_md()

        default_profile = Modulemd.Profile(name="default")
        for pkg in self.pkgs:
            default_profile.add_rpm(pkg)

        self.mmd.add_profile(default_profile)

    def _write_container_yaml(self, output_fname):
        pkg = self.pkgs[0]
        command = pkg

        container_yaml = dedent(f'''\
            compose:
                modules:
                    - {pkg}:master
            flatpak:
                # Derived from the project's domain name
                id: org.example.MyApp
                branch: stable
                # Binary to execute to run the app
                command: {command}
                tags: []
                # Not sandboxed. See 'man flatpak-build-finish'
                finish-args: >
                    --filesystem=host
                    --share=ipc
                    --socket=x11
                    --socket=wayland
                    --socket=session-bus
            ''')

        with open(output_fname, 'w') as f:
            f.write(container_yaml)

        print('Generated container specification: %r. Please edit appropriately.' % output_fname)

    def run(self, output_modulemd, output_containerspec, force=False):
        if output_modulemd is None:
            pkg = self.pkgs[0]
            output_modulemd = pkg + '.yaml'
        if output_containerspec is None:
            output_containerspec = 'container.yaml'

        if not force:
            if os.path.exists(output_modulemd):
                raise click.ClickException(f"{output_modulemd} exists. Pass"
                                           f" --force to overwrite.")
            if os.path.exists(output_containerspec):
                raise click.ClickException(f"{output_containerspec} exists."
                                           f" Pass --force to overwrite.")

        super().run(output_modulemd)

        self._write_container_yaml(output_containerspec)


def do_flatpak_report(pkgs, quiet=False):
    if not quiet:
        print("Initializing", file=sys.stderr)
    runtime_packages = _get_runtime_packages()
    pool = _depchase.make_pool("x86_64")

    packages = {}
    flatpaks = {}

    for p in runtime_packages:
        packages[p] = {
            'name': p,
            'runtime': True,
            'used_by': []
        }

    for pkg in pkgs:
        if not quiet:
            print("Calculating deps for", pkg, file=sys.stderr)
        all_needed_packages = _depchase.ensure_installable(pool, [pkg], hints=sorted(runtime_packages))
        extra_packages = all_needed_packages - runtime_packages - set([pkg])
        used_runtime_packages = all_needed_packages - extra_packages - set([pkg])
        flatpaks[pkg] = {
            'runtime': sorted(used_runtime_packages),
            'extra': sorted(extra_packages)
        }
        for p in all_needed_packages:
            if p == pkg:
                continue
            data = packages.get(p, None)
            if data is None:
                data = {
                    'name': p,
                    'runtime': False,
                    'used_by': []
                }
                packages[p] = data
            data['used_by'].append(pkg)

    json.dump({
        'packages': packages,
        'flatpaks': flatpaks
    }, sys.stdout, indent=4)
