import logging
import re

import gi
gi.require_version('Modulemd', '1.0')
from gi.repository import Modulemd

from . import _depchase, _repodata
from .util import rpm_name_only

def _unwrap_description(desc):
    def replace(m):
        if len(m.group(0)) == 1:
            return ' '
        else:
            return m.group(0)[1:]

    return re.sub(r'\n+', replace, desc)

def _categorise_deps(all_rpm_deps):
    module_deps = set()
    remaining_rpm_deps = set()
    for pkgname in all_rpm_deps:
        modname = _repodata.get_module_for_rpm(pkgname)
        if modname is not None:
            module_deps.add(modname)
        else:
            remaining_rpm_deps.add(pkgname)
    return module_deps, remaining_rpm_deps

class ModuleGenerator(object):

    def __init__(self, pkgs):
        self.pkgs = pkgs
        self.mmd = Modulemd.Module(mdversion=2)
        self._pool = _depchase.make_pool("x86_64")
        self.rpm_branch = None

    def _calculate_dependencies(self):
        pkgs = self.pkgs
        pool = self._pool
        self.api_srpms = {rpm_name_only(_depchase.get_srpm_for_rpm(pool, dep)) for dep in pkgs}
        run_deps = _depchase.ensure_installable(pool, pkgs)
        module_run_deps, rpm_run_deps = _categorise_deps(run_deps)
        # The platform module provides any other runtime dependencies - we expect we'll
        # always depend on it, but check for the heck of it.
        if len(rpm_run_deps) > 0:
            module_run_deps.add('platform')
        # We default to 'all available streams' for all dependencies we've found. The module
        # author may then edit this to exclude streams or restrict to specific streams.
        self.module_run_deps = { m: [] for m in module_run_deps }
        self.run_srpms = set()

    def _save_module_md(self, output_fname):
        """
        Function writes modulemd to either stdout or the given filename
        :return:
        """
        if output_fname is not None:
            self.mmd.dump(output_fname)
            print('Generated modulemd file: %r' % output_fname)
        else:
            print(self.mmd.dumps())
        return True

    def _update_module_md(self):
        """
        Function updates modulemd file with dependencies
        are information taken from SPEC file.
        :return:
        """
        self.mmd.peek_module_licenses().add("MIT")

        # FIXME: This and:
        # https://fedoraproject.org/wiki/Module:Guidelines?rd=Fedora_Packaging_Guidelines_for_Modules
        #
        # should be improved so that it's clear what type of information should be in the
        # resulting summary and description.
        if len(self.pkgs) == 1:
            # If we're generating a module for a single package, use the summary and
            # description from the package. This isn't fully correct, but gives the
            # module maintainer a starting place to edit from.
            metadata = _depchase.get_rpm_metadata(self._pool, self.pkgs[0])
            self.mmd.props.summary = metadata['summary'].strip()
            self.mmd.props.description = _unwrap_description(metadata['description'].strip())
        else:
            self.mmd.props.summary = f"Generated module for {self.pkgs}"
            self.mmd.props.description = "Module auto-generated by fedmod"


        # Declare the public API
        for pkg in self.api_srpms:
            self.mmd.peek_rpm_api().add(pkg)

            component = Modulemd.ComponentRpm(name=pkg,
                                              rationale="Package in api",
                                              buildorder=self._get_build_order(pkg))
            if self.rpm_branch is not None:
                component.props.ref = self.rpm_branch
            self.mmd.add_rpm_component(component)

        # Declare module level dependencies
        dependencies = Modulemd.Dependencies()
        for modname, streams in self.module_run_deps.items():
            dependencies.add_buildrequires(modname, streams)
            dependencies.add_requires(modname, streams)
        self.mmd.add_dependencies(dependencies)

        for pkg in self.run_srpms:
            component = Modulemd.ComponentRpm(name=pkg,
                                              rationale="Runtime dependencies",
                                              buildorder=self._get_build_order(pkg))
            if self.rpm_branch is not None:
                component.props.ref = self.rpm_branch
            self.mmd.add_rpm_component(component)

        # TODO: Always set content licenses appropriately

        # TODO: Emit something for non-empty self.unresolved_build_rpms
        #       rather than relying solely on the warnings emitted on stderr


    def _get_build_order(self, pkg):
        if self.mmd.peek_rpm_api().contains(pkg):
            return 10
        else:
            return 0

    def run(self, output_fname):
        self._calculate_dependencies()
        self._update_module_md()
        self._save_module_md(output_fname)
