# -*- coding: utf-8 -*-
#
# modulemd_linter - Linter for ModuleMD files
#
# Copyright © 2017, 2018 Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author:
# Nils Philippsen <nils@redhat.com>

import fnmatch
import inspect
from itertools import chain
from shutil import get_terminal_size

from click.formatting import wrap_text

import yaml


TEXT_INDENT = "    "


class guidelines_link(object):
    """Decorate a linter method with a link to the guidelines (can be a full
    URL or just an anchor on the main wiki page)."""

    guidelines_base_url = "https://fedoraproject.org/wiki/Module:Guidelines"

    def __init__(self, link=None):
        self.link = link

    def __call__(self, fn):
        if not self.link or "://" in self.link:
            fn.lint_guidelines_url = self.link
        else:
            anchor = self.link
            if anchor.startswith("#"):
                anchor = anchor[1:]
            fn.lint_guidelines_url = "{}#{}".format(self.guidelines_base_url,
                                                    anchor)
        return fn


class LintLevel(object):
    """Decorator to specify a 'severity' level for a linter method."""

    all_levels = {}
    max_level = -1

    # We store this here because every linter method needs to be decorated with
    # a lint level.
    lint_order = 0

    def __init__(self, level_name, level):
        assert(level not in self.all_levels)
        self.level_name = level_name
        self.level = level
        self.all_levels[level_name] = level
        LintLevel.max_level = max(LintLevel.max_level, level)

    def __call__(self, fn):
        fn.lint_level = self.level
        fn.lint_level_name = self.level_name
        fn.lint_order = LintLevel.lint_order
        if not hasattr(fn, 'lint_prerequisites'):
            fn.lint_prerequisites = set()
        LintLevel.lint_order += 1
        return fn


option = LintLevel("option", -1)
info = LintLevel("info", 0)
warning = LintLevel("warning", 1)
error = LintLevel("error", 2)


class detail(object):
    """Decorate a linter method with details about what needs to be fixed."""

    def __init__(self, detail):
        self.detail = detail

    def __call__(self, fn):
        fn.lint_detail = self.detail
        return fn


class prerequisite_base(object):
    """Base class for the @prerequisite, @prerequisite_for decorators."""

    attr_name = None

    def __init__(self, *names):
        assert len(names)

        # Allow specification with and without "lint_" prefix
        self.names = set(x if x.startswith("lint_") else "lint_" + x
                         for x in names)

    def __call__(self, fn):
        try:
            getattr(fn, self.attr_name).update(self.names)
        except AttributeError:
            setattr(fn, self.attr_name, self.names)
        return fn


class prerequisite(prerequisite_base):
    """Decorate a linter method with what checks are a prerequisite so as to
    not overwhelm a user with subsequent errors."""

    attr_name = 'lint_prerequisites'


class prerequisite_for(prerequisite_base):
    """Decorate a linter method with what checks have it as a prerequisite.

    Accepts globs."""

    attr_name = 'lint_prerequisites_for'


class LintCheckFailed(Exception):
    pass


class ModuleMDLinter(object):
    """Linter for modulemd YAML files or data.

    Contains one method for each check to perform and scaffolding. The names of
    the methods performing checks begin with 'lint_', they do nothing if the
    check succeeds but throw a LintCheckFailed exception otherwise. An easy way
    to do this is using the available check*() methods."""

    # PyYAML decodes numerical strings as int or float
    yaml_scalar_types = (str, int, float)

    # 'str' for any scalar type
    components_keys_types = {
        'rationale': str,
        'buildorder': int,
        'repository': str,
        'ref': str,
        'cache': str,
        'arches': [str],
        'multilib': [str],
    }

    recognized_rpm_keys = set(components_keys_types)

    recognized_module_keys = {'rationale', 'buildorder', 'repository', 'ref'}

    def __init__(self, modulemd_path=None, modulemd_str=None):
        if bool(modulemd_path) == bool(modulemd_str):
            raise ValueError(
                "specify exactly one of modulemd_path or modulemd_str")

        self.modulemd_path = modulemd_path
        self.modulemd_str = modulemd_str

    def lint(self, min_level=None):
        # ensure that min_level describes a valid level
        if min_level is None:
            min_level = 0
        elif isinstance(min_level, int):
            assert 0 <= min_level <= LintLevel.max_level
        else:
            assert min_level in LintLevel.all_levels
            min_level = LintLevel.all_levels[min_level]

        lint_method_names = [x for x in dir(self) if x.startswith("lint_")]

        # ensure all lint methods are properly decorated
        assert all(
            -1 <= getattr(getattr(self, x), "lint_level", -2)
               <= LintLevel.max_level
            for x in lint_method_names), (
                "All lint_*() methods must be decorated with one of the"
                " @option, @info, @warning, @error decorators.")

        lint_methods = sorted(
            (y for y in (getattr(self, x) for x in lint_method_names)
             if callable(y) and (y.lint_level == -1
                                 or y.lint_level >= min_level)),
            key=lambda x: x.lint_order)

        names_to_methods = {m.__name__: m for m in lint_methods}

        # add names of methods decorated with @prerequisite_for to the list of
        # prerequisites of matching methods
        for method in lint_methods:
            name = method.__name__
            for glob in getattr(method, 'lint_prerequisites_for', ()):
                matching_names = fnmatch.filter(names_to_methods, glob)
                for n in matching_names:
                    # don't apply to decorated method
                    if n == name:
                        continue
                    names_to_methods[n].lint_prerequisites.add(name)

        # ensure all lint methods have a docstring
        assert all(
            getattr(x, '__doc__', None)
            for x in lint_methods if x.lint_level >= 0), (
                """All non-option lint_*() methods must have a docstring.""")

        flagged = []
        failed_or_skipped = set()

        for method in lint_methods:
            skip_check = False

            for prerequisite in getattr(method, 'lint_prerequisites', ()):
                if prerequisite in failed_or_skipped:
                    skip_check = True
                    failed_or_skipped.add(method.__name__)
                    break

            if skip_check:
                continue

            try:
                method()
            except LintCheckFailed as e:
                if method.lint_level >= 0:
                    try:
                        lint_detail = method.lint_detail
                    except AttributeError:
                        lint_detail = None
                    flagged.append((method.lint_level_name,
                                    inspect.getdoc(method), lint_detail,
                                    method.lint_guidelines_url, e))
                failed_or_skipped.add(method.__name__)

        width, height = get_terminal_size()

        for i, (lvl, text, detail, url, exc) in enumerate(flagged):
            print("[{lvl}] {url}".format(lvl=lvl.upper(),
                                         url=url or ""))

            print()
            print(wrap_text(text=text, width=width - 2,
                            initial_indent=TEXT_INDENT,
                            subsequent_indent=TEXT_INDENT,
                            preserve_paragraphs=True))

            if detail:
                print()
                print(TEXT_INDENT + "DETAIL:")
                print(wrap_text(text=detail, width=width - 2,
                                initial_indent=TEXT_INDENT,
                                subsequent_indent=TEXT_INDENT,
                                preserve_paragraphs=True))

            if i < len(flagged) - 1:
                print()

        return any(lambda x: x[0] not in ('option', 'info') for x in flagged)

    def check_true(self, expr):
        """Check that a value is True or true-like."""
        if not expr:
            raise LintCheckFailed()

    def check(self, expr):
        """Synonym for check_true()."""
        self.check_true(expr)

    def check_false(self, expr):
        """Check that a value is not True or true-like."""
        self.check_true(not expr)

    def check_is_scalar(self, *values):
        """Check that 0..n values are of a scalar type."""
        for value in values:
            self.check_true(isinstance(value, self.yaml_scalar_types))

    def check_is_list(self, *values):
        """Check that 0..n values are lists."""
        for value in values:
            self.check_true(isinstance(value, list))

    def check_is_dict(self, *values):
        """Check that 0..n values are of a dictionary type."""
        for value in values:
            self.check_true(isinstance(value, dict))

    # lint checks below

    @error
    @guidelines_link("http://yaml.org/spec/current.html#id2513364")
    @detail("YAML files must be encoded in UTF-16 or UTF-8. However, we only"
            " support UTF-8 encodings currently.")
    def lint_document_is_unicode(self):
        """A YAML processor must support the UTF-16 and UTF-8 character
        encodings."""
        if not self.modulemd_str:
            try:
                with open(self.modulemd_path, "r") as mmdfile:
                    self.modulemd_str = mmdfile.read()
            except UnicodeDecodeError:
                raise LintCheckFailed()
        else:
            self.check(isinstance(self.modulemd_str, str))

    @error
    @guidelines_link()
    @detail("The document is not a YAML file.")
    @prerequisite('document_is_unicode')
    @prerequisite_for('*')
    def lint_document_is_yaml(self):
        """Each module is defined by a single YAML file and comprises of a
        number of key-value pairs describing the module's properties and
        components it contains."""
        try:
            self.yaml_doc = yaml.safe_load(self.modulemd_str)
        except:
            raise LintCheckFailed()

    @error
    @guidelines_link()
    @detail("The document is not a ModuleMD YAML file.")
    @prerequisite_for('*')
    def lint_document_is_modulemd_yaml(self):
        """Each module is defined by a single YAML file and comprises of a
        number of key-value pairs describing the module's properties and
        components it contains."""
        try:
            self.check(self.yaml_doc['document'] == 'modulemd')
            self.check(str(self.yaml_doc['version']) in ('1', '2'))
            # We intentionally don't use the modulemd module for linting, its
            # parser isn't well-suited for continuing after problems and we
            # want to be able to list as many as possible.
            self.mmd = self.yaml_doc['data']
        except:
            raise LintCheckFailed()

    @option
    @prerequisite_for('v1_*')
    def lint_modulemd_is_version_1(self):
        self.check(str(self.yaml_doc['version']) == '1')

    @option
    @prerequisite_for('v2_*')
    def lint_modulemd_is_version_2(self):
        self.check(str(self.yaml_doc['version']) == '2')

    @error
    @guidelines_link('#Summary_and_description')
    def lint_summary_exists(self):
        """Every module MUST include a short summary."""
        self.check(self.mmd.get('summary', "").strip())

    @error
    @guidelines_link('#Summary_and_description')
    @prerequisite('summary_exists')
    def lint_summary_one_sentence(self):
        """The module summary is a one sentence concise description of the
        module."""
        self.check("." not in self.mmd['summary'][:-1])

    @warning
    @guidelines_link('#Summary_and_description')
    @prerequisite('summary_exists')
    def lint_summary_no_trailing_period(self):
        """The summary SHOULD NOT end in a period."""
        self.check(not self.mmd['summary'].strip().endswith("."))

    @error
    @guidelines_link('#Summary_and_description')
    def lint_description_exists(self):
        """Every module MUST include a description."""
        self.check('description' in self.mmd
                   and self.mmd['description'].strip())

    @warning
    @guidelines_link('#Summary_and_description')
    @prerequisite('description_exists')
    def lint_description_ends_in_period(self):
        """The description SHOULD end in a period."""
        self.check(self.mmd['description'].strip().endswith("."))

    @error
    @guidelines_link('#Licensing')
    def lint_license_block_exists(self):
        """Every module MUST contain a license block."""
        self.check('license' in self.mmd)

    @error
    @guidelines_link('#Licensing')
    @detail("The license block must be a dictionary (key/value).")
    @prerequisite('license_block_exists')
    def lint_license_block_dict(self):
        """Every module MUST contain a license block."""
        self.check_is_dict(self.mmd['license'])

    @warning
    @guidelines_link('#Licensing')
    @detail("The license block must be a dictionary (key/value). Recognized"
            " keys are 'module' and 'content'.")
    @prerequisite('license_block_dict')
    def lint_license_block_keys(self):
        """Every module MUST contain a license block."""
        self.check(set(self.mmd['license']) <= {'module', 'content'})

    @error
    @guidelines_link('#Licensing')
    @detail("All license blocks ('module' and 'content') must be lists.")
    @prerequisite('license_block_dict')
    def lint_license_blocks_are_lists(self):
        """Every module MUST contain a license block and declare a list of
        the module's licenses."""
        self.check_is_list(*self.mmd['license'].values())

    @error
    @guidelines_link('#Licensing')
    @detail("Each license must be a scalar value.")
    @prerequisite('license_blocks_are_lists')
    def lint_licenses_scalar(self):
        """Every module MUST contain a license block and declare a list of
        the module's licenses."""
        self.check_is_scalar(*chain(*self.mmd['license'].values()))

    @error
    @guidelines_link('#Licensing')
    @detail("The module license block must exist.")
    @prerequisite('license_block_dict')
    def lint_license_module_block_exists(self):
        """Every module MUST contain a license block and declare a list of
        the module's licenses."""
        self.check('module' in self.mmd['license'])

    @error
    @guidelines_link('#Short_overview')
    def lint_no_manual_name_stream_version_context_arch(self):
        """Module packagers MUST NOT define values for name, stream, version,
        context or architecture manually but should rather expect the Module
        Build Service to do it for them."""
        self.check_false(
            set(('name', 'stream', 'version', 'context', 'arch'))
            & set(self.mmd))

    @error
    @guidelines_link('#Short_overview')
    def lint_no_manual_eol(self):
        """Module packagers MUST NOT define the EOL in the modulemd but should
        define it in other infrastructure tooling."""
        self.check_false('eol' in self.mmd)

    @warning
    @prerequisite_for('*_dependencies_*')
    @guidelines_link('#Dependencies')
    @detail("The 'dependencies' block is missing.")
    def lint_dependencies_exists(self):
        """Modules MAY depend on other modules. These module relationships are
        listed in the dependencies block. However, none of modules are
        buildable without the platform module hence packagers are required to
        specify it."""
        self.check(self.mmd.get('dependencies'))

    @error
    @guidelines_link('#Dependencies')
    @detail("The dependencies block must be a dictionary (key/value).")
    @prerequisite_for('v1_dependencies_*')
    def lint_v1_dependencies_dict(self):
        """Modules MAY depend on other modules. These module relationships are
        listed in the dependencies block."""
        self.check_is_dict(self.mmd['dependencies'])

    @warning
    @guidelines_link('#Dependencies')
    @detail("The module doesn't require 'platform'.")
    def lint_v1_dependencies_platform_reqd(self):
        """Modules MAY depend on other modules. These module relationships are
        listed in the dependencies block. However, none of modules are
        buildable without the platform module hence packagers are required to
        specify it."""
        self.check(all('platform' in x for x in
                       self.mmd['dependencies'].values()))

    @error
    @guidelines_link('#Dependencies')
    @detail("The dependencies block must be a list.")
    @prerequisite_for('v2_dependencies_*')
    def lint_v2_dependencies_block_is_list(self):
        """Modules MAY depend on other modules. These module relationships are
        listed in the dependencies block. Unlike in the first version of the
        format, the dependencies block is a list of dictionaries."""
        self.check_is_list(self.mmd['dependencies'])

    @error
    @guidelines_link('#Dependencies')
    @detail("Each list item in the dependencies block must be a dict.")
    @prerequisite_for('v2_dependencies_*')
    def lint_v2_dependencies_block_is_list_of_dicts(self):
        """Modules MAY depend on other modules. These module relationships are
        listed in the dependencies block. Unlike in the first version of the
        format, the dependencies block is a list of dictionaries."""
        self.check_is_dict(*self.mmd['dependencies'])

    @error
    @guidelines_link('#Dependencies')
    @prerequisite_for('v2_dependencies_items_contain_platform')
    @detail("buildrequires must exist")
    def lint_v2_dependencies_items_have_buildrequires(self):
        """Modules MAY depend on other modules. These module relationships are
        listed in the dependencies block. However, none of modules are
        buildable without platform module hence packagers are required to
        specify it."""
        for block in self.mmd['dependencies']:
            self.check('buildrequires' in block)

    @error
    @guidelines_link('#Dependencies')
    @prerequisite_for('v2_dependencies_items_contain_platform')
    @detail("requires must exist")
    def lint_v2_dependencies_items_have_requires(self):
        """Modules MAY depend on other modules. These module relationships are
        listed in the dependencies block. However, none of modules are
        buildable without platform module hence packagers are required to
        specify it."""
        for block in self.mmd['dependencies']:
            self.check('requires' in block)

    @error
    @guidelines_link('#Dependencies')
    @detail("The streams of module dependencies must be specified as lists.")
    def lint_v2_dependencies_modstreams_must_be_lists(self):
        """Modules MAY depend on other modules. These module relationships are
        listed in the dependencies block."""
        for block in self.mmd['dependencies']:
            for mod_names_streams in block.values():
                self.check_is_list(*(v for v in mod_names_streams.values()))

    @error
    @guidelines_link('#Dependencies')
    @detail("Each stream of a module dependency must be scalar.")
    @prerequisite('v2_dependencies_modstreams_must_be_lists')
    def lint_v2_dependencies_modstreams_must_be_lists_of_scalars(self):
        """Modules MAY depend on other modules. These module relationships are
        listed in the dependencies block."""
        for block in self.mmd['dependencies']:
            for mod_names_streams in block.values():
                for streams in mod_names_streams.values():
                    self.check_is_scalar(*streams)

    @error
    @guidelines_link('#Dependencies')
    @detail("modules must depend on platform")
    def lint_v2_dependencies_items_contain_platform(self):
        """Modules MAY depend on other modules. These module relationships are
        listed in the dependencies block. However, none of modules are
        buildable without platform module hence packagers are required to
        specify it."""
        for block in self.mmd['dependencies']:
            for items in block.values():
                self.check('platform' in items)

    @error
    @guidelines_link('#Dependencies')
    @prerequisite_for('v1_dependencies_*')
    @detail("dependencies must be a dictionary")
    def lint_v1_dependencies_must_be_dict(self):
        """Modules MAY depend on other modules. These module relationships are
        listed in the dependencies block."""
        self.check_is_dict(self.mmd['dependencies'])

    @error
    @guidelines_link('#Dependencies')
    @prerequisite_for('v1_dependencies_buildrequires_*')
    @detail("buildrequires must exist")
    def lint_v1_dependencies_buildrequires_exists(self):
        """Modules MAY depend on other modules. These module relationships are
        listed in the dependencies block."""
        self.check(self.mmd['dependencies'].get('buildrequires'))

    @error
    @guidelines_link('#Dependencies')
    @detail("The 'buildrequires' block beneath 'dependencies' must be a"
            " dictionary (key/value).")
    @prerequisite('v1_dependencies_buildrequires_exists')
    def lint_v1_dependencies_buildrequires_dict(self):
        """Modules MAY depend on other modules. These module relationships are
        listed in the dependencies block."""
        self.check_is_dict(self.mmd['dependencies']['buildrequires'])

    @error
    @guidelines_link('#Dependencies')
    @detail("The 'buildrequires' dictionary must associate module name keys"
            " with stream name values.")
    @prerequisite('v1_dependencies_buildrequires_dict')
    def lint_v1_dependencies_buildrequires_items(self):
        """Modules MAY depend on other modules. These module relationships are
        listed in the dependencies block. Dependencies are expressed using
        module names and their stream names."""
        self.check_is_scalar(
            *self.mmd['dependencies']['buildrequires'].values())

    @option
    @prerequisite_for('v1_dependencies_requires_*')
    def lint_v1_dependencies_requires_exists(self):
        self.check(self.mmd['dependencies'].get('requires'))

    @error
    @guidelines_link('#Dependencies')
    @detail("The 'requires' block beneath 'dependencies' must be a dictionary"
            " (key/value).")
    @prerequisite('v1_dependencies_dict')
    def lint_v1_dependencies_requires_dict(self):
        """Modules MAY depend on other modules. These module relationships are
        listed in the dependencies block. Dependencies are expressed using
        module names and their stream names."""
        self.check_is_dict(self.mmd['dependencies']['requires'])

    @error
    @guidelines_link('#Dependencies')
    @detail("The 'requires' dictionary must associate module name keys with"
            " stream name values.")
    @prerequisite('v1_dependencies_requires_dict')
    def lint_v1_dependencies_requires_items(self):
        """Modules MAY depend on other modules. These module relationships are
        listed in the dependencies block. Dependencies are expressed using
        module names and their stream names."""
        self.check_is_scalar(*self.mmd['dependencies']['requires'].values())

    @option
    @prerequisite_for('xmd_*')
    def lint_xmd_exists(self):
        self.check(self.mmd.get('xmd'))

    @error
    @guidelines_link('#Extensible_module_metadata_block')
    @detail("The 'xmd' block must be a dictionary (key/value).")
    def lint_xmd_dict(self):
        """Modules MAY also contain an extensible metadata block, a list of
        vendor-defined key-value pairs."""
        self.check_is_dict(self.mmd['xmd'])

    @option
    @prerequisite_for('references_*')
    def lint_references_exists(self):
        self.check(self.mmd.get('references'))

    @error
    @guidelines_link('#References')
    @detail("The 'references' block must be a dictionary (key/value).")
    @prerequisite_for('references_*')
    def lint_references_dict(self):
        """Modules MAY define links referencing various upstream resources,
        such as community website, project documentation or upstream bug
        tracker."""
        self.check_is_dict(self.mmd['references'])

    @warning
    @guidelines_link('#References')
    @detail("The 'references' block must be a dictionary (key/value)."
            " Recognized keys are 'community', 'documentation' and 'tracker'.")
    def lint_references_keys(self):
        """Modules MAY define links referencing various upstream resources,
        such as community website, project documentation or upstream bug
        tracker."""
        self.check(set(self.mmd['references'])
                   <= {'community', 'documentation', 'tracker'})

    @error
    @guidelines_link('#References')
    @detail("The references must be scalar values (URLs).")
    def lint_references_urls(self):
        """Modules MAY define links referencing various upstream resources,
        such as community website, project documentation or upstream bug
        tracker."""
        self.check_is_scalar(*self.mmd['references'].values())

    @option
    @prerequisite_for('profiles_*')
    def lint_profiles_exists(self):
        self.check(self.mmd.get('profiles'))

    @error
    @guidelines_link('#Profiles')
    @detail("The 'profiles' block must be a dictionary (key/value).")
    def lint_profiles_dict(self):
        """The module author MAY define lists of packages that would be
        installed by default, and a minimum, when the module is enabled and the
        particular profile is selected."""
        self.check_is_dict(self.mmd['profiles'])

    @info
    @guidelines_link('#Profiles')
    @detail("Modules that are supposed to be user-installable should in most"
            " cases have at least one profile.")
    @prerequisite('profiles_dict')
    def lint_profiles_not_empty(self):
        """The module author MAY define lists of packages that would be
        installed by default, and a minimum, when the module is enabled and the
        particular profile is selected."""
        self.check(len(self.mmd['profiles']))

    @info
    @guidelines_link('#Profiles')
    @detail("Modules that are supposed to be user-installable should in most"
            " cases have a 'default' profile.")
    @prerequisite('profiles_dict')
    def lint_profiles_has_default(self):
        """The module author MAY define lists of packages that would be
        installed by default, and a minimum, when the module is enabled and the
        particular profile is selected."""
        self.check('default' in self.mmd['profiles'])

    @error
    @guidelines_link('#Profiles')
    @detail("The 'profiles' block must contain dictionary (key/value) blocks.")
    @prerequisite('profiles_dict')
    def lint_profiles_blocks_are_dicts(self):
        """The module author MAY define lists of packages that would be
        installed by default, and a minimum, when the module is enabled and the
        particular profile is selected."""
        self.check_is_dict(*self.mmd['profiles'].values())

    @warning
    @guidelines_link('#API')
    def lint_api_exists(self):
        """Every module SHOULD define its public API."""
        self.check(self.mmd.get('api'))

    @error
    @guidelines_link('#API')
    @detail("The 'api' block must be a dictionary (key/value).")
    @prerequisite('api_exists')
    def lint_api_dict(self):
        """Every module SHOULD define its public API."""
        self.check_is_dict(self.mmd['api'])

    @warning
    @guidelines_link('#API')
    @detail("The 'api' block should have the 'rpms' key (and only that).")
    @prerequisite('api_dict')
    def lint_api_keys(self):
        """Every module SHOULD define its public API. Currently the only
        supported type of API are binary RPM packages."""
        self.check(set(self.mmd['api']) == {'rpms'})

    @error
    @guidelines_link('#API')
    @detail("All blocks beneath 'api' must be lists.")
    @prerequisite('api_dict')
    def lint_api_blocks_are_lists(self):
        """Every module SHOULD define its public API."""
        self.check_is_list(*self.mmd['api'].values())

    @warning
    @guidelines_link('#API')
    @detail("The list of API RPM packages is empty.")
    @prerequisite('api_keys', 'api_blocks_are_lists')
    def lint_api_rpms_empty(self):
        """Every module SHOULD define its public API."""
        self.check(len(self.mmd['api']['rpms']) > 0)

    @error
    @guidelines_link('#API')
    @detail("Every block beneath 'api' must be a list of scalar items.")
    @prerequisite('api_blocks_are_lists')
    def lint_api_blocks_items_are_scalar(self):
        """Every module SHOULD define its public API."""
        for block in self.mmd['api'].values():
            self.check_is_scalar(*block)

    @option
    @prerequisite_for('filter_*')
    def lint_filter_exists(self):
        self.check(self.mmd.get('filter'))

    @error
    @guidelines_link('#Filters')
    @detail("The 'filter' block must be a dictionary (key/value).")
    def lint_filter_dict(self):
        """Module filters define lists of components or other content that
        should not be part of the resulting, composed module deliverable."""
        self.check_is_dict(self.mmd['filter'])

    @warning
    @guidelines_link('#Filters')
    @detail("The 'filter' block should have the 'rpms' key (and only that).")
    @prerequisite('filter_dict')
    def lint_filter_rpms(self):
        """Module filters define lists of components or other content that
        should not be part of the resulting, composed module deliverable.
        Currently the only supported type of filter are binary RPM packages."""
        self.check(set(self.mmd['filter']) == {'rpms'})

    @error
    @guidelines_link('#Filters')
    @detail("All blocks beneath 'filter' must be lists.")
    @prerequisite('filter_dict')
    def lint_filter_blocks_are_lists(self):
        """Module filters define lists of components or other content that
        should not be part of the resulting, composed module deliverable."""
        self.check_is_list(*self.mmd['filter'].values())

    @option
    @prerequisite_for('filter_rpms_*')
    @prerequisite('filter_dict')
    def lint_filter_rpms_exists(self):
        self.check(self.mmd['filter'].get('rpms'))

    @warning
    @guidelines_link('#Filters')
    @detail("The list of filtered RPM packages is empty.")
    @prerequisite('filter_blocks_are_lists')
    def lint_filter_rpms_empty(self):
        """Module filters define lists of components or other content that
        should not be part of the resulting, composed module deliverable.
        Currently the only supported type of filter are binary RPM packages."""
        self.check(len(self.mmd['filter']['rpms']) > 0)

    @error
    @guidelines_link('#Filters')
    @detail("Every block beneath 'filter' must be a list of scalar items.")
    @prerequisite('filter_blocks_are_lists')
    def lint_filter_blocks_items_are_scalar(self):
        """Module filters define lists of components or other content that
        should not be part of the resulting, composed module deliverable."""
        for block in self.mmd['filter'].values():
            self.check_is_scalar(*block)

    @option
    @prerequisite_for('components_*')
    def lint_components_exists(self):
        self.check(self.mmd.get('components'))

    @error
    @guidelines_link('#Short_overview')
    @detail("The 'components' block must be a dictionary (key/value).")
    def lint_components_dict(self):
        """Modules MAY, and most modules do contain a components block defining
        the module's content."""
        self.check_is_dict(self.mmd['components'])

    @warning
    @guidelines_link('#Short_overview')
    @detail("The 'components' block only recognizes 'rpms' and 'modules' as"
            " keys.")
    @prerequisite('components_dict')
    def lint_components_keys(self):
        """Modules MAY, and most modules do contain a components block defining
        the module's content. RPM packages and other modules are the only
        currently supported content types."""
        self.check(set(self.mmd['components']) <= {'rpms', 'modules'})

    @error
    @guidelines_link('#Short_overview')
    @detail("All blocks beneath 'components' must be dictionaries"
            " (key/value).")
    @prerequisite('components_dict')
    def lint_components_blocks_dict(self):
        """Modules MAY, and most modules do contain a components block defining
        the module's content."""
        self.check_is_dict(*self.mmd['components'].values())

    @error
    @guidelines_link('#Short_overview')
    @detail("All blocks two levels beneath 'components' must be dictionaries"
            " (key/value).")
    @prerequisite('components_blocks_dict')
    def lint_components_subblocks_dict(self):
        """Modules MAY, and most modules do contain a components block defining
        the module's content."""
        self.check_is_dict(*(subblock
                             for block in self.mmd['components'].values()
                             for subblock in block.values()))

    @warning
    @guidelines_link('#Short_overview')
    @detail("One or more blocks beneath 'components' are empty.")
    @prerequisite('components_blocks_dict')
    def lint_components_blocks_empty(self):
        """Modules MAY, and most modules do contain a components block defining
        the module's content."""
        self.check(all(self.mmd['components'].values()))

    @warning
    @guidelines_link('#Short_overview')
    @detail("One or more blocks two levels beneath 'components' are empty.")
    @prerequisite('components_blocks_dict')
    def lint_components_subblocks_empty(self):
        """Modules MAY, and most modules do contain a components block defining
        the module's content."""
        self.check(all(subblock
                       for block in self.mmd['components'].values()
                       for subblock in block.values()))

    @error
    @guidelines_link('#Short_overview')
    @prerequisite('components_subblocks_dict')
    def lint_components_blocks_mandatory_keys(self):
        """Every component MUST declare why it was added to the module, its
        'rationale'."""
        for block in self.mmd['components'].values():
            for subblock in self.mmd['components']['rpms'].values():
                self.check('rationale' in subblock)
                self.check_is_scalar(subblock['rationale'])

    @error
    @guidelines_link('#Short_overview')
    @detail("Illegal type of component key(s): 'buildorder' must be an"
            " integer, 'arches' and 'multilib' a list of strings and all"
            " others strings.")
    @prerequisite('components_subblocks_dict')
    def lint_components_blocks_keys_types(self):
        """Modules MAY, and most modules do contain a components block defining
        the module's content."""
        for block in self.mmd['components'].values():
            for comp_md in block.values():
                for k, v in comp_md.items():
                    t = self.components_keys_types.get(k)
                    if not t:
                        # unknown key
                        continue

                    if isinstance(t, list):
                        self.check_is_list(v)
                        t = t[0]
                    else:
                        v = [v]

                    if issubclass(t, str):
                        t = self.yaml_scalar_types

                    for value in v:
                        self.check(isinstance(value, t))

    @option
    @prerequisite_for('components_rpms_*')
    def lint_components_rpms_exists(self):
        self.check(self.mmd['components'].get('rpms'))

    @warning
    @guidelines_link('#Short_overview')
    @detail("Recognized keys are: {}".format(
        ", ".join("'{}'".format(k) for k in recognized_rpm_keys)))
    @prerequisite('components_subblocks_dict')
    def lint_components_rpms_recognized_keys(self):
        """Module RPM content is defined in the 'rpms' block of
        'components'."""
        self.check(all(set(b) <= self.recognized_rpm_keys
                       for b in self.mmd['components']['rpms'].values()))

    @option
    @prerequisite_for('components_modules_*')
    def lint_components_modules_exists(self):
        self.check(self.mmd['components'].get('modules'))

    @warning
    @guidelines_link('#Short_overview')
    @detail("Recognized keys are: {}".format(
        ", ".join("'{}'".format(k) for k in recognized_module_keys)))
    @prerequisite('components_subblocks_dict')
    def lint_components_modules_recognized_keys(self):
        """Including modules as content is defined in the 'modules' block of
        'components'."""
        self.check(
            all(set(b) <= self.recognized_module_keys
                for b in self.mmd['components']['modules'].values()))
